
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import de.unkrig.commons.file.contentstransformation.ContentsTransformer;
import de.unkrig.commons.io.MarkableFileInputStream;

/**
 * A {@link FileTransformer} that transforms a file be feeding its contents through a {@link ContentsTransformer}.
 */
public
class FileContentsTransformer implements FileTransformer {

    private final ContentsTransformer contentsTransformer;

    public
    FileContentsTransformer(ContentsTransformer contentsTransformer) {
        this.contentsTransformer = contentsTransformer;
    }

    /**
     * Opens the {@code in} file for reading, opens the {@code out} file for writing, reads the contents, feeds it
     * through the given {@link ContentsTransformer}, and writes it to the {@code out} file.
     */
    @Override public void
    transform(File in, File out) throws IOException {
        if (in.equals(out)) throw new AssertionError("IN and out must not be identical: " + in);

        InputStream is = new MarkableFileInputStream(in);
        try {

            OutputStream os = new FileOutputStream(out);
            try {
                this.contentsTransformer.transform(in.getPath(), is, os);
                os.close();
            } catch (IOException ioe) {
                try { os.close(); } catch (IOException ioe2) {}
                out.delete();

                throw wrap("Transforming file '" + in + "' into '" + out + "'", ioe);
            } catch (RuntimeException re) {
                try { os.close(); } catch (IOException ioe2) {}
                out.delete();

                throw wrap("Transforming file '" + in + "' into '" + out + "'", re);
            } finally {
                try { os.close(); } catch (IOException ioe) {}
            }
            is.close();
        } finally {
            try { is.close(); } catch (IOException ioe) {}
        }

        if (!out.setLastModified(in.lastModified())) {
            throw new IOException("Could not set modification time of '" + out + "'");
        }
    }
}
