
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.lang.protocol;

import de.unkrig.commons.nullanalysis.Nullable;

public
class ProducerUtil {

    private ProducerUtil() {}

    /**
     * Converts a {@link Producer} (the "source") into a different {@link Producer} (the "target"), which is possible
     * iff the source's element type is a subclass of the target's element type.
     *
     * @param <TT> The target producer's element type
     * @param <ST> The source producer's element type
     */
    @SuppressWarnings({ "rawtypes", "unchecked" })
    public static <TT, ST extends TT> Producer<TT>
    widen(final Producer<ST> source) {
        return (Producer) source;
    }

    /**
     * Converts a {@link ProducerWhichThrows} (the "source") into a different {@link ProducerWhichThrows} (the
     * "target"), which is possible iff:
     * <ul>
     *   <li>The source's element type is a subclass of the target's element type, and
     *   <li>The source's exception is a subclass of the target's exception.
     * </ul>
     *
     * @param <TT> The target producer's element type
     * @param <ST> The source producer's element type
     * @param <TE> The target producer's exception
     * @param <SE> The source producer's exception
     */
    @SuppressWarnings({ "rawtypes", "unchecked" })
    public static <TT, ST extends TT, TE extends Throwable, SE extends TE> ProducerWhichThrows<TT, TE>
    widen(final ProducerWhichThrows<ST, SE> source) {
        return (ProducerWhichThrows) source;
    }

    /**
     * Converts a {@link Producer Producer&lt;ST>} (the "source") into a {@link ProducerWhichThrows
     * ProducerWhichThrows&lt;TT, TE>} (the "target"), which is possible iff the source's element type is a subclass of
     * the target's element type.
     *
     * @param <TT> The target producer's element type
     * @param <ST> The source producer's element type
     * @param <TE> The target producer's exception
     */
    public static <TT, ST extends TT, TE extends Throwable> ProducerWhichThrows<TT, TE>
    cast(final Producer<ST> source) {
        return new ProducerWhichThrows<TT, TE>() {

            @Override public @Nullable TT
            produce() { return source.produce(); }
        };
    }

    /**
     * Converts a {@link ProducerWhichThrows ProducerWhichThrows&lt;ST, RuntimeException>} (the "source") into a
     * {@link Producer Producer&lt;TT>} (the "target"), which is possible iff the source's element type is a subclass
     * of the target's element type.
     *
     * @param <TT> The target producer's element type
     * @param <ST> The source producer's element type
     */
    public static <TT, ST extends TT> Producer<TT>
    cast(final ProducerWhichThrows<ST, RuntimeException> source) {
        return new Producer<TT>() {

            @Override public @Nullable TT
            produce() { return source.produce(); }
        };
    }

    /**
     * @return A {@link Producer} that produced the given {@code elements}
     */
    public static <T> Producer<T>
    fromElements(final T... elements) {
    	return new Producer<T>() {

    		int idx = 0;

			@Override public @Nullable T
			produce() {
				 return this.idx == elements.length ? null : elements[this.idx++];
			}
		};
    }
}
